#include <algorithm>
#include <msp/fs/dir.h>
#include <msp/fs/stat.h>
#include <msp/fs/utils.h>
#include <msp/io/print.h>
#include <msp/strings/lexicalcast.h>
#include "booleanevaluator.h"
#include "builder.h"
#include "component.h"
#include "sourcepackage.h"

using namespace std;
using namespace Msp;

Component::Component(SourcePackage &p, const string &n):
	package(p),
	name(n),
	install(false),
	deflt(true)
{ }

void Component::prepare()
{
	for(Package::Requirements::const_iterator i=requires.begin(); i!=requires.end(); ++i)
		(*i)->prepare();
}

void Component::create_build_info()
{
	BuildInfo final_build_info;

	const Package::Requirements &pkg_reqs = package.get_required_packages();
	Package::Requirements direct_reqs = requires;
	direct_reqs.insert(direct_reqs.end(), pkg_reqs.begin(), pkg_reqs.end());

	Package::Requirements all_reqs = direct_reqs;
	for(Package::Requirements::iterator i=all_reqs.begin(); i!=all_reqs.end(); ++i)
	{
		BuildInfo::UpdateLevel level = BuildInfo::CHAINED;
		if(find(direct_reqs.begin(), direct_reqs.end(), *i)!=direct_reqs.end())
			level = BuildInfo::DEPENDENCY;
		final_build_info.update_from((*i)->get_exported_build_info(), level);

		const Package::Requirements &reqs = (*i)->get_required_packages();
		for(Package::Requirements::const_iterator j=reqs.begin(); j!=reqs.end(); ++j)
			if(find(all_reqs.begin(), all_reqs.end(), *j)==all_reqs.end())
				all_reqs.push_back(*j);
	}

	final_build_info.update_from(package.get_build_info());
	final_build_info.update_from(build_info);
	build_info = final_build_info;

	for(BuildInfo::PathList::iterator i=build_info.incpath.begin(); i!=build_info.incpath.end(); ++i)
		*i = (package.get_source_directory() / *i).str();
	for(BuildInfo::PathList::iterator i=build_info.libpath.begin(); i!=build_info.libpath.end(); ++i)
		*i = (package.get_source_directory() / *i).str();
}

BuildInfo Component::get_build_info_for_path(const FS::Path &path) const
{
	// XXX Cache these and check that the directories actually exist before adding them
	BuildInfo binfo = build_info;
	if(!overlays.empty())
	{
		FS::Path dir = FS::dirname(path);
		string last = FS::basename(dir);
		for(OverlayList::const_iterator i=overlays.begin(); i!=overlays.end(); ++i)
			if(last==*i)
			{
				dir = FS::dirname(dir);
				break;
			}

		for(SourceList::const_iterator i=sources.begin(); i!=sources.end(); ++i)
			if(dir==*i)
			{
				binfo.local_incpath.push_back(dir);
				for(OverlayList::const_iterator j=overlays.begin(); j!=overlays.end(); ++j)
					binfo.local_incpath.push_back(*i/ *j);
			}
	}
	return binfo;
}

Component::SourceList Component::collect_source_files() const
{
	SourceList files;
	for(SourceList::const_iterator i=sources.begin(); i!=sources.end(); ++i)
	{
		FS::Path path(*i);
		if(FS::is_dir(path))
		{
			SourceList dirs;
			dirs.push_back(path);
			for(OverlayList::const_iterator j=overlays.begin(); j!=overlays.end(); ++j)
			{
				FS::Path opath = path / *j;
				if(FS::is_dir(opath))
					dirs.push_back(opath);
			}
			for(SourceList::const_iterator j=dirs.begin(); j!=dirs.end(); ++j)
			{
				package.get_builder().get_logger().log("files", format("Traversing %s", *j));
				list<string> sfiles = list_files(*j);
				for(list<string>::iterator k=sfiles.begin(); k!=sfiles.end(); ++k)
					files.push_back(*j / *k);
			}
		}
		else
		{
			files.push_back(path);
			for(OverlayList::const_iterator j=overlays.begin(); j!=overlays.end(); ++j)
			{
				FS::Path opath = FS::dirname(path)/ *j/FS::basename(path);
				if(FS::is_reg(opath))
					files.push_back(opath);
			}
		}
	}

	return files;
}


Component::Loader::Loader(Component &c):
	DataFile::ObjectLoader<Component>(c)
{
	add("if_arch",         &Loader::if_arch);
	add("if_feature",      &Loader::if_feature);
	add("overlay",         &Loader::overlay);
	add("source",          &Loader::source);
	add("install",         &Component::install);
	add("install_map",     &Loader::install_map);
	add("build_info",      &Loader::build_info);
	add("require",         &Loader::require);
	add("default",         &Component::deflt);
}

void Component::Loader::build_info()
{
	load_sub(obj.build_info);
}

void Component::Loader::if_arch(const string &cond)
{
	BooleanEvaluator eval(sigc::hide<1>(sigc::mem_fun(&obj.package.get_builder().get_current_arch(), &Architecture::match_name)), false);
	bool match = eval.evaluate(cond);
	obj.package.get_builder().get_logger().log("configure",
		format("%s/%s: arch %s %smatched", obj.package.get_name(), obj.name, cond, (match ? "" : "not ")));
	if(match)
		load_sub_with(*this);
}

void Component::Loader::if_feature(const string &cond)
{
	BooleanEvaluator eval(sigc::mem_fun(&obj.package, &SourcePackage::match_feature));
	bool match = eval.evaluate(cond);
	obj.package.get_builder().get_logger().log("configure",
		format("%s/%s: feature %s %smatched", obj.package.get_name(), obj.name, cond, (match ? "" : "not ")));
	if(match)
		load_sub_with(*this);
}

void Component::Loader::install_map()
{
	load_sub(obj.install_map, obj.package.get_source_directory());
}

void Component::Loader::overlay(const string &o)
{
	obj.overlays.push_back(o);
}

void Component::Loader::require(const string &n)
{
	Package *req = obj.package.get_builder().get_package_manager().find_package(n);
	if(req)
		obj.requires.push_back(req);
	else
		obj.problems.push_back(format("Required package %s not found", n));
}

void Component::Loader::source(const string &s)
{
	obj.sources.push_back((obj.package.get_source_directory()/s).str());
}
