#include <msp/fs/stat.h>
#include <msp/fs/utils.h>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include <msp/time/utils.h>
#include "builder.h"
#include "filetarget.h"
#include "sourcepackage.h"
#include "task.h"
#include "tool.h"

using namespace std;
using namespace Msp;

FileTarget::FileTarget(Builder &b, const FS::Path &a):
	Target(b, generate_name(b, 0, a)),
	path(a)
{
	init(0);
}

FileTarget::FileTarget(Builder &b, const SourcePackage &p, const FS::Path &a):
	Target(b, generate_name(b, &p, a)),
	path(a)
{
	init(&p);
}

void FileTarget::init(const SourcePackage *p)
{
	size = 0;
	package = p;
	nested_build_sig = false;
	arch_in_build_sig = false;

	builder.get_vfs().register_path(path, this);

	stat();
}

void FileTarget::stat()
{
	if(FS::Stat st = FS::lstat(path))
	{
		mtime = st.get_modify_time();
		size = st.get_size();
	}
}

string FileTarget::generate_name(Builder &builder, const SourcePackage *pkg, const FS::Path &path)
{
	if(pkg && FS::descendant_depth(path, pkg->get_source_directory())>=0)
	{
		FS::Path relpath = FS::relative(path, pkg->get_source_directory());
		return format("<%s>%s", pkg->get_name(), relpath.str().substr(1));
	}
	else if(FS::descendant_depth(path, builder.get_prefix())>=0)
	{
		FS::Path relpath = FS::relative(path, builder.get_prefix());
		return "<prefix>"+relpath.str().substr(1);
	}

	return path.str();
}

void FileTarget::touch()
{
	mtime = Time::now();
	modified();
	signal_bubble_rebuild.emit();
}

void FileTarget::check_rebuild()
{
	if(!tool || needs_rebuild())
		return;

	if(!mtime)
		mark_rebuild("Does not exist");
	else
	{
		for(Dependencies::iterator i=depends.begin(); (i!=depends.end() && !needs_rebuild()); ++i)
		{
			FileTarget *ft = dynamic_cast<FileTarget *>(*i);
			if(ft && ft->get_mtime()>mtime)
				mark_rebuild((*i)->get_name()+" has changed");
			else if((*i)->needs_rebuild())
				mark_rebuild((*i)->get_name()+" needs rebuilding");
		}
	}

	if(!needs_rebuild())
	{
		// Some side effects might not exist
		for(Dependencies::iterator i=side_effects.begin(); (i!=side_effects.end() && !needs_rebuild()); ++i)
			if((*i)->needs_rebuild())
				mark_rebuild((*i)->get_name()+" needs rebuilding");
	}

	if(!needs_rebuild() && package)
	{
		if(package->get_config().get_mtime()>mtime)
			mark_rebuild("Package options changed");

		if(tool->get_executable())
		{
			string build_sig = create_build_signature();
			if(package->get_cache().has_key(this, "build_sig"))
			{
				if(package->get_cache().get_value(this, "build_sig")!=build_sig)
					mark_rebuild("Build signature changed");
			}
		}
	}
}

string FileTarget::create_build_signature() const
{
	if(!package)
		return string();

	const BuildInfo &binfo = (component ? component->get_build_info() : package->get_build_info());
	list<string> sigs;
	if(nested_build_sig && component)
	{
		set<const Tool *> depend_tools;
		for(Dependencies::const_iterator i=depends.begin(); i!=depends.end(); ++i)
			if((*i)->get_component()==component && (*i)->get_tool())
				depend_tools.insert((*i)->get_tool());

		for(set<const Tool *>::const_iterator i=depend_tools.begin(); i!=depend_tools.end(); ++i)
			sigs.push_back((*i)->create_build_signature(binfo));
		sigs.sort();
		sigs.push_front(tool->create_build_signature(binfo));
	}
	else
		sigs.push_back(tool->create_build_signature(binfo));

	if(arch_in_build_sig)
		if(const Architecture *arch = tool->get_architecture())
			sigs.push_front(arch->get_name());

	return join(sigs.begin(), sigs.end(), ";");
}

void FileTarget::build(Task &task)
{
	task.add_file(path);
	task.set_unlink(true);
}

void FileTarget::build_finished(bool success)
{
	if(success)
	{
		stat();
		if(package)
		{
			string build_sig = create_build_signature();
			if(!build_sig.empty())
				package->get_cache().set_value(this, "build_sig", build_sig);
		}
	}

	Target::build_finished(success);
}

void FileTarget::clean()
{
	if(mtime)
	{
		FS::unlink(path);
		mtime = Time::TimeStamp();
		size = 0;
		check_rebuild();
	}
}
