#include <cstdlib>
#include <msp/fs/utils.h>
#include <msp/strings/format.h>
#include "builder.h"
#include "component.h"
#include "exportdefinitions.h"
#include "externaltask.h"
#include "importlibrary.h"
#include "installedfile.h"
#include "mingwdlltool.h"
#include "objectfile.h"
#include "sharedlibrary.h"
#include "sourcepackage.h"

using namespace std;
using namespace Msp;

MingwDllTool::MingwDllTool(Builder &b, const Architecture &a):
	Tool(b, a, "DLL")
{
	set_command("dlltool", true);
}

Target *MingwDllTool::create_target(const list<Target *> &sources, const string &)
{
	if(sources.size()!=1)
		throw invalid_argument("MingwDllTool::create_target");
	SharedLibrary &shlib = dynamic_cast<SharedLibrary &>(*sources.front());

	list<ObjectFile *> objs;
	const Target::Dependencies &depends = shlib.get_dependencies();
	for(Target::Dependencies::const_iterator i=depends.begin(); i!=depends.end(); ++i)
		if(ObjectFile *obj = dynamic_cast<ObjectFile *>(*i))
			objs.push_back(obj);

	ExportDefinitions *exp = new ExportDefinitions(builder, *shlib.get_component(), objs);
	exp->set_tool(*this);

	ImportLibrary *imp = new ImportLibrary(builder, *shlib.get_component(), shlib, *exp);
	imp->set_tool(*this);

	return imp;
}

Target *MingwDllTool::create_install(Target &target) const
{
	if(ImportLibrary *imp = dynamic_cast<ImportLibrary *>(&target))
	{
		Tool &copy = builder.get_toolchain().get_tool("CP");
		InstalledFile *inst_tgt = dynamic_cast<InstalledFile *>(copy.create_target(target));
		string link_name = format("lib%s.dll.a", imp->get_shared_library()->get_libname());
		if(link_name!=FS::basename(inst_tgt->get_path()))
			inst_tgt->set_symlink(link_name);
		return inst_tgt;
	}
	else
		return 0;
}

Task *MingwDllTool::run(const Target &target) const
{
	const ImportLibrary *imp = dynamic_cast<const ImportLibrary *>(&target);
	const ExportDefinitions *exp = 0;
	if(imp)
		exp = &dynamic_cast<const ExportDefinitions &>(*imp->get_dependencies().front());
	else
		exp = dynamic_cast<const ExportDefinitions *>(&target);
	if(!imp && !exp)
		throw invalid_argument("MingwDllTool::run");

	vector<string> argv;
	argv.push_back(executable->get_path().str());

	/* dlltool is stupid and puts temporary files in the working directory by
	default */
	argv.push_back("--temp-prefix");
	char random[8];
	for(unsigned i=0; i<8; ++i)
		random[i] = 'a'+(rand()%26);
	argv.push_back(string("/tmp/")+string(random, 8));

	const Component &comp = *target.get_component();
	FS::Path work_dir = comp.get_package().get_source_directory();

	if(imp)
	{
		const SharedLibrary &shlib = *imp->get_shared_library();

		argv.push_back("-d");
		argv.push_back(relative(exp->get_path(), work_dir).str());

		argv.push_back("-D");
		if(shlib.get_install_filename().empty())
			argv.push_back(FS::basename(shlib.get_path()));
		else
			argv.push_back(shlib.get_install_filename());

		argv.push_back("-l");
		argv.push_back(relative(imp->get_path(), work_dir).str());
	}
	else
	{
		const Target::Dependencies &depends = exp->get_dependencies();
		for(Target::Dependencies::const_iterator i=depends.begin(); i!=depends.end(); ++i)
		{
			if(ObjectFile *obj = dynamic_cast<ObjectFile *>(*i))
				argv.push_back(relative(obj->get_path(), work_dir).str());
		}

		// XXX Should use dllexport, but that has some other problems to solve
		argv.push_back("--export-all-symbols");

		argv.push_back("-z");
		argv.push_back(relative(exp->get_path(), work_dir).str());
	}

	return new ExternalTask(argv, work_dir);
}
